from mediafile.utils import sc_decode, sc_encode


class StorageStyle:
    """A strategy for storing a value for a certain tag format (or set
    of tag formats). This basic StorageStyle describes simple 1:1
    mapping from raw values to keys in a Mutagen file object; subclasses
    describe more sophisticated translations or format-specific access
    strategies.

    MediaFile uses a StorageStyle via three methods: ``get()``,
    ``set()``, and ``delete()``. It passes a Mutagen file object to
    each.

    Internally, the StorageStyle implements ``get()`` and ``set()``
    using two steps that may be overridden by subtypes. To get a value,
    the StorageStyle first calls ``fetch()`` to retrieve the value
    corresponding to a key and then ``deserialize()`` to convert the raw
    Mutagen value to a consumable Python value. Similarly, to set a
    field, we call ``serialize()`` to encode the value and then
    ``store()`` to assign the result into the Mutagen object.

    Each StorageStyle type has a class-level `formats` attribute that is
    a list of strings indicating the formats that the style applies to.
    MediaFile only uses StorageStyles that apply to the correct type for
    a given audio file.
    """

    formats = [
        "FLAC",
        "OggOpus",
        "OggTheora",
        "OggSpeex",
        "OggVorbis",
        "OggFlac",
        "APEv2File",
        "WavPack",
        "Musepack",
        "MonkeysAudio",
    ]
    """List of mutagen classes the StorageStyle can handle.
    """

    def __init__(self, key, as_type=str, suffix=None, float_places=2, read_only=False):
        """Create a basic storage strategy. Parameters:

        - `key`: The key on the Mutagen file object used to access the
          field's data.
        - `as_type`: The Python type that the value is stored as
          internally (`unicode`, `int`, `bool`, or `bytes`).
        - `suffix`: When `as_type` is a string type, append this before
          storing the value.
        - `float_places`: When the value is a floating-point number and
          encoded as a string, the number of digits to store after the
          decimal point.
        - `read_only`: When true, writing to this field is disabled.
           Primary use case is so wrongly named fields can be addressed
           in a graceful manner. This does not block the delete method.

        """
        self.key = key
        self.as_type = as_type
        self.suffix = suffix
        self.float_places = float_places
        self.read_only = read_only

        # Convert suffix to correct string type.
        if self.suffix and self.as_type is str and not isinstance(self.suffix, str):
            self.suffix = self.suffix.decode("utf-8")

    # Getter.

    def get(self, mutagen_file):
        """Get the value for the field using this style."""
        return self.deserialize(self.fetch(mutagen_file))

    def fetch(self, mutagen_file):
        """Retrieve the raw value of for this tag from the Mutagen file
        object.
        """
        try:
            return mutagen_file[self.key][0]
        except (KeyError, IndexError):
            return None

    def deserialize(self, mutagen_value):
        """Given a raw value stored on a Mutagen object, decode and
        return the represented value.
        """
        if (
            self.suffix
            and isinstance(mutagen_value, str)
            and mutagen_value.endswith(self.suffix)
        ):
            return mutagen_value[: -len(self.suffix)]
        else:
            return mutagen_value

    # Setter.

    def set(self, mutagen_file, value):
        """Assign the value for the field using this style."""
        self.store(mutagen_file, self.serialize(value))

    def store(self, mutagen_file, value):
        """Store a serialized value in the Mutagen file object."""
        mutagen_file[self.key] = [value]

    def serialize(self, value):
        """Convert the external Python value to a type that is suitable for
        storing in a Mutagen file object.
        """
        if isinstance(value, float) and self.as_type is str:
            value = "{0:.{1}f}".format(value, self.float_places)
            value = self.as_type(value)
        elif self.as_type is str:
            if isinstance(value, bool):
                # Store bools as 1/0 instead of True/False.
                value = str(int(bool(value)))
            elif isinstance(value, bytes):
                value = value.decode("utf-8", "ignore")
            else:
                value = str(value)
        else:
            value = self.as_type(value)

        if self.suffix:
            value += self.suffix

        return value

    def delete(self, mutagen_file):
        """Remove the tag from the file."""
        if self.key in mutagen_file:
            del mutagen_file[self.key]


class ListStorageStyle(StorageStyle):
    """Abstract storage style that provides access to lists.

    The ListMediaField descriptor uses a ListStorageStyle via two
    methods: ``get_list()`` and ``set_list()``. It passes a Mutagen file
    object to each.

    Subclasses may overwrite ``fetch`` and ``store``.  ``fetch`` must
    return a (possibly empty) list or `None` if the tag does not exist.
    ``store`` receives a serialized list of values as the second argument.

    The `serialize` and `deserialize` methods (from the base
    `StorageStyle`) are still called with individual values. This class
    handles packing and unpacking the values into lists.
    """

    def get(self, mutagen_file):
        """Get the first value in the field's value list."""
        values = self.get_list(mutagen_file)
        if values is None:
            return None

        try:
            return values[0]
        except IndexError:
            return None

    def get_list(self, mutagen_file):
        """Get a list of all values for the field using this style."""
        raw_values = self.fetch(mutagen_file)
        if raw_values is None:
            return None

        return [self.deserialize(item) for item in raw_values]

    def fetch(self, mutagen_file):
        """Get the list of raw (serialized) values."""
        try:
            return mutagen_file[self.key]
        except KeyError:
            return None

    def set(self, mutagen_file, value):
        """Set an individual value as the only value for the field using
        this style.
        """
        if value is None:
            self.store(mutagen_file, None)
        else:
            self.set_list(mutagen_file, [value])

    def set_list(self, mutagen_file, values):
        """Set all values for the field using this style. `values`
        should be an iterable.
        """
        if values is None:
            self.delete(mutagen_file)
        else:
            self.store(mutagen_file, [self.serialize(value) for value in values])

    def store(self, mutagen_file, values):
        """Set the list of all raw (serialized) values for this field."""
        mutagen_file[self.key] = values


class SoundCheckStorageStyleMixin:
    """A mixin for storage styles that read and write iTunes SoundCheck
    analysis values. The object must have an `index` field that
    indicates which half of the gain/peak pair---0 or 1---the field
    represents.
    """

    def get(self, mutagen_file):
        data = self.fetch(mutagen_file)
        if data is not None:
            return sc_decode(data)[self.index]

    def set(self, mutagen_file, value):
        data = self.fetch(mutagen_file)
        if data is None:
            gain_peak = [0, 0]
        else:
            gain_peak = list(sc_decode(data))
        gain_peak[self.index] = value or 0
        data = self.serialize(sc_encode(*gain_peak))
        self.store(mutagen_file, data)



import re


def safe_cast(out_type, val):
    """Try to covert val to out_type but never raise an exception.

    If the value does not exist, return None. Or, if the value
    can't be converted, then a sensible default value is returned.
    out_type should be bool, int, or unicode; otherwise, the value
    is just passed through.
    """
    if val is None:
        return None

    if out_type is int:
        if isinstance(val, int) or isinstance(val, float):
            # Just a number.
            return int(val)
        else:
            # Process any other type as a string.
            if isinstance(val, bytes):
                val = val.decode("utf-8", "ignore")
            elif not isinstance(val, str):
                val = str(val)
            # Get a number from the front of the string.
            match = re.match(r"[\+-]?[0-9]+", val.strip())
            return int(match.group(0)) if match else 0

    elif out_type is bool:
        try:
            # Should work for strings, bools, ints:
            return bool(int(val))
        except ValueError:
            return False

    elif out_type is str:
        if isinstance(val, bytes):
            return val.decode("utf-8", "ignore")
        elif isinstance(val, str):
            return val
        else:
            return str(val)

    elif out_type is float:
        if isinstance(val, int) or isinstance(val, float):
            return float(val)
        else:
            if isinstance(val, bytes):
                val = val.decode("utf-8", "ignore")
            else:
                val = str(val)
            match = re.match(r"[\+-]?([0-9]+\.?[0-9]*|[0-9]*\.[0-9]+)", val.strip())
            if match:
                val = match.group(0)
                if val:
                    return float(val)
            return 0.0

    else:
        return val



from mediafile.constants import ImageType
from mediafile.utils import Image

from .base import ListStorageStyle


class APEv2ImageStorageStyle(ListStorageStyle):
    """Store images in APEv2 tags. Values are `Image` objects."""

    formats = ["APEv2File", "WavPack", "Musepack", "MonkeysAudio", "OptimFROG"]

    TAG_NAMES = {
        ImageType.other: "Cover Art (other)",
        ImageType.icon: "Cover Art (icon)",
        ImageType.other_icon: "Cover Art (other icon)",
        ImageType.front: "Cover Art (front)",
        ImageType.back: "Cover Art (back)",
        ImageType.leaflet: "Cover Art (leaflet)",
        ImageType.media: "Cover Art (media)",
        ImageType.lead_artist: "Cover Art (lead)",
        ImageType.artist: "Cover Art (artist)",
        ImageType.conductor: "Cover Art (conductor)",
        ImageType.group: "Cover Art (band)",
        ImageType.composer: "Cover Art (composer)",
        ImageType.lyricist: "Cover Art (lyricist)",
        ImageType.recording_location: "Cover Art (studio)",
        ImageType.recording_session: "Cover Art (recording)",
        ImageType.performance: "Cover Art (performance)",
        ImageType.screen_capture: "Cover Art (movie scene)",
        ImageType.fish: "Cover Art (colored fish)",
        ImageType.illustration: "Cover Art (illustration)",
        ImageType.artist_logo: "Cover Art (band logo)",
        ImageType.publisher_logo: "Cover Art (publisher logo)",
    }

    def __init__(self):
        super().__init__(key="")

    def fetch(self, mutagen_file):
        images = []
        for cover_type, cover_tag in self.TAG_NAMES.items():
            try:
                frame = mutagen_file[cover_tag]
                text_delimiter_index = frame.value.find(b"\x00")
                if text_delimiter_index > 0:
                    comment = frame.value[0:text_delimiter_index]
                    comment = comment.decode("utf-8", "replace")
                else:
                    comment = None
                image_data = frame.value[text_delimiter_index + 1 :]
                images.append(Image(data=image_data, type=cover_type, desc=comment))
            except KeyError:
                pass

        return images

    def set_list(self, mutagen_file, values):
        self.delete(mutagen_file)

        for image in values:
            image_type = image.type or ImageType.other
            comment = image.desc or ""
            image_data = comment.encode("utf-8") + b"\x00" + image.data
            cover_tag = self.TAG_NAMES[image_type]
            mutagen_file[cover_tag] = image_data

    def delete(self, mutagen_file):
        """Remove all images from the file."""
        for cover_tag in self.TAG_NAMES.values():
            try:
                del mutagen_file[cover_tag]
            except KeyError:
                pass



# Cover art and other images.

import logging

import filetype

from mediafile.constants import ImageType

log = logging.getLogger(__name__)


def image_mime_type(data):
    """Return the MIME type of the image data (a bytestring)."""
    return filetype.guess_mime(data)


def image_extension(data):
    ext = filetype.guess_extension(data)
    # imghdr returned "tiff", so we should keep returning it with filetype.
    return ext if ext != "tif" else "tiff"


class Image:
    """Structure representing image data and metadata that can be
    stored and retrieved from tags.

    The structure has four properties.
    * ``data``  The binary data of the image
    * ``desc``  An optional description of the image
    * ``type``  An instance of `ImageType` indicating the kind of image
    * ``mime_type`` Read-only property that contains the mime type of
                    the binary data
    """

    def __init__(self, data, desc=None, type=None):
        assert isinstance(data, bytes)
        if desc is not None:
            assert isinstance(desc, str)
        self.data = data
        self.desc = desc
        if isinstance(type, int):
            try:
                type = list(ImageType)[type]
            except IndexError:
                log.debug("ignoring unknown image type index %s", type)
                type = ImageType.other
        self.type = type

    @property
    def mime_type(self):
        if self.data:
            return image_mime_type(self.data)

    @property
    def type_index(self):
        if self.type is None:
            # This method is used when a tag format requires the type
            # index to be set, so we return "other" as the default value.
            return 0
        return self.type.value



import mutagen.flac

from mediafile.utils import Image

from .base import ListStorageStyle


class FlacImageStorageStyle(ListStorageStyle):
    """Converts between ``mutagen.flac.Picture`` and ``Image`` instances."""

    formats = ["FLAC"]

    def __init__(self):
        super().__init__(key="")

    def fetch(self, mutagen_file):
        return mutagen_file.pictures

    def deserialize(self, flac_picture):
        return Image(
            data=flac_picture.data, desc=flac_picture.desc, type=flac_picture.type
        )

    def store(self, mutagen_file, pictures):
        """``pictures`` is a list of mutagen.flac.Picture instances."""
        mutagen_file.clear_pictures()
        for pic in pictures:
            mutagen_file.add_picture(pic)

    def serialize(self, image):
        """Turn a Image into a mutagen.flac.Picture."""
        pic = mutagen.flac.Picture()
        pic.data = image.data
        pic.type = image.type_index
        pic.mime = image.mime_type
        pic.desc = image.desc or ""
        return pic

    def delete(self, mutagen_file):
        """Remove all images from the file."""
        mutagen_file.clear_pictures()



# This file is part of MediaFile.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Handles low-level interfacing for files' tags. Wraps Mutagen to
automatically detect file types and provide a unified interface for a
useful subset of music files' tags.

Usage:

    >>> f = MediaFile('Lucy.mp3')
    >>> f.title
    u'Lucy in the Sky with Diamonds'
    >>> f.artist = 'The Beatles'
    >>> f.save()

A field will always return a reasonable value of the correct type, even
if no tag is present. If no value is available, the value will be false
(e.g., zero or the empty string).

Internally ``MediaFile`` uses ``MediaField`` descriptors to access the
data from the tags. In turn ``MediaField`` uses a number of
``StorageStyle`` strategies to handle format specific logic.
"""

import logging
import os
import re

import mutagen
import mutagen.mp3

from .constants import TYPES, ImageType
from .exceptions import FileTypeError, MutagenError, UnreadableFileError
from .fields import (
    CoverArtField,
    DateField,
    DateItemField,
    ImageListField,
    ListMediaField,
    MediaField,
    QNumberField,
)
from .storage import (
    APEv2ImageStorageStyle,
    ASFImageStorageStyle,
    ASFStorageStyle,
    FlacImageStorageStyle,
    ListStorageStyle,
    MP3DescStorageStyle,
    MP3ImageStorageStyle,
    MP3ListDescStorageStyle,
    MP3ListPeopleStorageStyle,
    MP3ListStorageStyle,
    MP3PeopleStorageStyle,
    MP3SlashPackStorageStyle,
    MP3SoundCheckStorageStyle,
    MP3StorageStyle,
    MP3UFIDStorageStyle,
    MP4BoolStorageStyle,
    MP4ImageStorageStyle,
    MP4ListStorageStyle,
    MP4SoundCheckStorageStyle,
    MP4StorageStyle,
    MP4TupleStorageStyle,
    SoundCheckStorageStyleMixin,
    StorageStyle,
    VorbisImageStorageStyle,
)
from .utils import (
    Image,
    image_extension,
    image_mime_type,
    loadfile,
    mutagen_call,
    update_filething,
)

__all__ = [
    "TYPES",
    "APEv2ImageStorageStyle",
    "ASFImageStorageStyle",
    "ASFStorageStyle",
    "CoverArtField",
    "DateField",
    "DateItemField",
    "FileTypeError",
    "FlacImageStorageStyle",
    "Image",
    "ImageListField",
    "ImageType",
    "ListMediaField",
    "ListStorageStyle",
    "MP3DescStorageStyle",
    "MP3ImageStorageStyle",
    "MP3ListDescStorageStyle",
    "MP3ListStorageStyle",
    "MP3PeopleStorageStyle",
    "MP3SlashPackStorageStyle",
    "MP3SoundCheckStorageStyle",
    "MP3StorageStyle",
    "MP3UFIDStorageStyle",
    "MP4BoolStorageStyle",
    "MP4ImageStorageStyle",
    "MP4ListStorageStyle",
    "MP4SoundCheckStorageStyle",
    "MP4StorageStyle",
    "MP4TupleStorageStyle",
    "MediaField",
    "MediaFile",
    "MutagenError",
    "QNumberField",
    "SoundCheckStorageStyleMixin",
    "StorageStyle",
    "UnreadableFileError",
    "VorbisImageStorageStyle",
    "image_extension",
    "image_mime_type",
    "loadfile",
    "mutagen_call",
]

log = logging.getLogger(__name__)


# MediaFile is a collection of fields.


class MediaFile:
    """Represents a multimedia file on disk and provides access to its
    metadata.
    """

    @loadfile()
    def __init__(self, filething, id3v23=False, raise_on_unsupported_wav=False):
        """Constructs a new `MediaFile` reflecting the provided file.

        `filething` can be a path to a file (i.e., a string) or a
        file-like object.

        May throw `UnreadableFileError`.

        By default, MP3 files are saved with ID3v2.4 tags. You can use
        the older ID3v2.3 standard by specifying the `id3v23` option.

        If `raise_on_unsupported_wav` is True, a `FileTypeError` is raised for WAV
        files containing non-PCM audio streams that cannot be tagged correctly.
        """
        self.filething = filething

        self.mgfile = mutagen_call("open", self.filename, mutagen.File, filething)

        if self.mgfile is None:
            # Mutagen couldn't guess the type
            raise FileTypeError(self.filename)
        elif type(self.mgfile).__name__ in ["M4A", "MP4"]:
            info = self.mgfile.info
            if info.codec and info.codec.startswith("alac"):
                self.type = "alac"
            else:
                self.type = "aac"
        elif type(self.mgfile).__name__ in ["ID3", "MP3"]:
            self.type = "mp3"
        elif type(self.mgfile).__name__ == "FLAC":
            self.type = "flac"
        elif type(self.mgfile).__name__ == "OggOpus":
            self.type = "opus"
        elif type(self.mgfile).__name__ == "OggVorbis":
            self.type = "ogg"
        elif type(self.mgfile).__name__ == "MonkeysAudio":
            self.type = "ape"
        elif type(self.mgfile).__name__ == "WavPack":
            self.type = "wv"
        elif type(self.mgfile).__name__ == "Musepack":
            self.type = "mpc"
        elif type(self.mgfile).__name__ == "ASF":
            self.type = "asf"
        elif type(self.mgfile).__name__ == "AIFF":
            self.type = "aiff"
        elif type(self.mgfile).__name__ == "DSF":
            self.type = "dsf"
        elif type(self.mgfile).__name__ == "WAVE":
            # WAV files can contain non-PCM audio streams that mutagen
            # cannot tag correctly. Rejects them with a clear error.
            _unsupported_wav_formats = {
                0x0002: "WAVE_FORMAT_ADPCM",
                0x0006: "WAVE_FORMAT_ALAW",
                0x0007: "WAVE_FORMAT_MULAW",
                0x0055: "WAVE_FORMAT_MPEGLAYER3",
            }
            audio_fmt = getattr(self.mgfile.info, "audio_format", 0x0001)
            if raise_on_unsupported_wav and audio_fmt in _unsupported_wav_formats:
                raise FileTypeError(self.filename, _unsupported_wav_formats[audio_fmt])
            self.type = "wav"
        else:
            raise FileTypeError(self.filename, type(self.mgfile).__name__)

        # Add a set of tags if it's missing.
        if self.mgfile.tags is None:
            self.mgfile.add_tags()

        # Set the ID3v2.3 flag only for MP3s.
        self.id3v23 = id3v23 and self.type == "mp3"

    @property
    def filename(self):
        """The name of the file.

        This is the path if this object was opened from the filesystem,
        or the name of the file-like object.
        """
        return self.filething.name

    @filename.setter
    def filename(self, val):
        """Silently skips setting filename.
        Workaround for `mutagen._util._openfile` setting instance's filename.
        """
        pass

    @property
    def path(self):
        """The path to the file.

        This is `None` if the data comes from a file-like object instead
        of a filesystem path.
        """
        return self.filething.filename

    @property
    def filesize(self):
        """The size (in bytes) of the underlying file."""
        if self.filething.filename:
            return os.path.getsize(self.filething.filename)
        if hasattr(self.filething.fileobj, "__len__"):
            return len(self.filething.fileobj)
        else:
            tell = self.filething.fileobj.tell()
            filesize = self.filething.fileobj.seek(0, 2)
            self.filething.fileobj.seek(tell)
            return filesize

    def save(self, **kwargs):
        """Write the object's tags back to the file.

        May throw `UnreadableFileError`. Accepts keyword arguments to be
        passed to Mutagen's `save` function.
        """
        # Possibly save the tags to ID3v2.3.
        if self.id3v23:
            id3 = self.mgfile
            if hasattr(id3, "tags"):
                # In case this is an MP3 object, not an ID3 object.
                id3 = id3.tags
            id3.update_to_v23()
            kwargs["v2_version"] = 3

        mutagen_call(
            "save",
            self.filename,
            self.mgfile.save,
            update_filething(self.filething),
            **kwargs,
        )

    def delete(self):
        """Remove the current metadata tag from the file. May
        throw `UnreadableFileError`.
        """
        mutagen_call(
            "delete",
            self.filename,
            self.mgfile.delete,
            update_filething(self.filething),
        )

    # Convenient access to the set of available fields.

    @classmethod
    def fields(cls):
        """Get the names of all writable properties that reflect
        metadata tags (i.e., those that are instances of
        :class:`MediaField`).
        """
        for property, descriptor in cls.__dict__.items():
            if isinstance(descriptor, MediaField):
                if isinstance(property, bytes):
                    # On Python 2, class field names are bytes. This method
                    # produces text strings.
                    yield property.decode("utf8", "ignore")
                else:
                    yield property

    @classmethod
    def _field_sort_name(cls, name):
        """Get a sort key for a field name that determines the order
        fields should be written in.

        Fields names are kept unchanged, unless they are instances of
        :class:`DateItemField`, in which case `year`, `month`, and `day`
        are replaced by `date0`, `date1`, and `date2`, respectively, to
        make them appear in that order.
        """
        if isinstance(cls.__dict__[name], DateItemField):
            name = re.sub("year", "date0", name)
            name = re.sub("month", "date1", name)
            name = re.sub("day", "date2", name)
        return name

    @classmethod
    def sorted_fields(cls):
        """Get the names of all writable metadata fields, sorted in the
        order that they should be written.

        This is a lexicographic order, except for instances of
        :class:`DateItemField`, which are sorted in year-month-day
        order.
        """
        yield from sorted(cls.fields(), key=cls._field_sort_name)

    @classmethod
    def readable_fields(cls):
        """Get all metadata fields: the writable ones from
        :meth:`fields` and also other audio properties.
        """
        yield from cls.fields()
        yield from (
            "length",
            "samplerate",
            "bitdepth",
            "bitrate",
            "bitrate_mode",
            "channels",
            "encoder_info",
            "encoder_settings",
            "format",
        )

    @classmethod
    def add_field(cls, name, descriptor):
        """Add a field to store custom tags.

        :param name: the name of the property the field is accessed
                     through. It must not already exist on this class.

        :param descriptor: an instance of :class:`MediaField`.
        """
        if not isinstance(descriptor, MediaField):
            raise ValueError(f"{descriptor} must be an instance of MediaField")
        if name in cls.__dict__:
            raise ValueError(f'property "{name}" already exists on MediaFile')
        setattr(cls, name, descriptor)

    def update(self, dict):
        """Set all field values from a dictionary.

        For any key in `dict` that is also a field to store tags the
        method retrieves the corresponding value from `dict` and updates
        the `MediaFile`. If a key has the value `None`, the
        corresponding property is deleted from the `MediaFile`.
        """
        for field in self.sorted_fields():
            if field in dict:
                if dict[field] is None:
                    delattr(self, field)
                else:
                    setattr(self, field, dict[field])

    def as_dict(self):
        """Get a dictionary with all writable properties that reflect
        metadata tags (i.e., those that are instances of
        :class:`MediaField`).
        """
        return {x: getattr(self, x) for x in self.fields()}

    # Field definitions.

    title = MediaField(
        MP3StorageStyle("TIT2"),
        MP4StorageStyle("\xa9nam"),
        StorageStyle("TITLE"),
        ASFStorageStyle("Title"),
    )
    artist = MediaField(
        MP3StorageStyle("TPE1"),
        MP4StorageStyle("\xa9ART"),
        StorageStyle("ARTIST"),
        ASFStorageStyle("Author"),
    )
    artists = ListMediaField(
        MP3ListDescStorageStyle(desc="ARTISTS"),
        MP4ListStorageStyle("----:com.apple.iTunes:ARTISTS"),
        ListStorageStyle("ARTISTS"),
        ASFStorageStyle("WM/ARTISTS"),
    )
    album = MediaField(
        MP3StorageStyle("TALB"),
        MP4StorageStyle("\xa9alb"),
        StorageStyle("ALBUM"),
        ASFStorageStyle("WM/AlbumTitle"),
    )
    genres = ListMediaField(
        MP3ListStorageStyle("TCON"),
        MP4ListStorageStyle("\xa9gen"),
        ListStorageStyle("GENRE"),
        ASFStorageStyle("WM/Genre"),
    )
    genre = genres.single_field()

    lyricists = ListMediaField(
        MP3ListStorageStyle("TEXT"),
        MP4ListStorageStyle("----:com.apple.iTunes:LYRICIST"),
        ListStorageStyle("LYRICIST"),
        ASFStorageStyle("WM/Writer"),
    )
    lyricist = lyricists.single_field()
    composers = ListMediaField(
        MP3ListStorageStyle("TCOM"),
        MP4ListStorageStyle("\xa9wrt"),
        ListStorageStyle("COMPOSER"),
        ASFStorageStyle("WM/Composer"),
    )
    composer = composers.single_field()
    composer_sort = MediaField(
        MP3StorageStyle("TSOC"),
        MP4StorageStyle("soco"),
        StorageStyle("COMPOSERSORT"),
        ASFStorageStyle("WM/Composersortorder"),
    )
    arrangers = ListMediaField(
        MP3ListPeopleStorageStyle("TIPL", involvement="arranger"),
        MP4ListStorageStyle("----:com.apple.iTunes:Arranger"),
        ListStorageStyle("ARRANGER"),
        ASFStorageStyle("beets/Arranger"),
    )
    arranger = arrangers.single_field()

    grouping = MediaField(
        MP3StorageStyle("TIT1"),
        MP4StorageStyle("\xa9grp"),
        StorageStyle("GROUPING"),
        ASFStorageStyle("WM/ContentGroupDescription"),
    )
    subtitle = MediaField(
        MP3StorageStyle("TIT3"),
        StorageStyle("SUBTITLE"),
        ASFStorageStyle("Subtitle"),
    )
    track = MediaField(
        MP3SlashPackStorageStyle("TRCK", pack_pos=0),
        MP4TupleStorageStyle("trkn", index=0),
        StorageStyle("TRACK"),
        StorageStyle("TRACKNUMBER"),
        ASFStorageStyle("WM/TrackNumber"),
        out_type=int,
    )
    tracktotal = MediaField(
        MP3SlashPackStorageStyle("TRCK", pack_pos=1),
        MP4TupleStorageStyle("trkn", index=1),
        StorageStyle("TRACKTOTAL"),
        StorageStyle("TRACKC"),
        StorageStyle("TOTALTRACKS"),
        ASFStorageStyle("TotalTracks"),
        out_type=int,
    )
    disc = MediaField(
        MP3SlashPackStorageStyle("TPOS", pack_pos=0),
        MP4TupleStorageStyle("disk", index=0),
        StorageStyle("DISC"),
        StorageStyle("DISCNUMBER"),
        ASFStorageStyle("WM/PartOfSet"),
        out_type=int,
    )
    disctotal = MediaField(
        MP3SlashPackStorageStyle("TPOS", pack_pos=1),
        MP4TupleStorageStyle("disk", index=1),
        StorageStyle("DISCTOTAL"),
        StorageStyle("DISCC"),
        StorageStyle("TOTALDISCS"),
        ASFStorageStyle("TotalDiscs"),
        out_type=int,
    )

    url = MediaField(
        MP3DescStorageStyle(key="WXXX", attr="url", multispec=False),
        MP4StorageStyle("\xa9url"),
        StorageStyle("URL"),
        ASFStorageStyle("WM/URL"),
    )
    lyrics = MediaField(
        MP3DescStorageStyle(key="USLT", multispec=False),
        MP4StorageStyle("\xa9lyr"),
        StorageStyle("LYRICS"),
        ASFStorageStyle("WM/Lyrics"),
    )
    comments = MediaField(
        MP3DescStorageStyle(key="COMM"),
        MP4StorageStyle("\xa9cmt"),
        StorageStyle("DESCRIPTION"),
        StorageStyle("COMMENT"),
        ASFStorageStyle("WM/Comments"),
        ASFStorageStyle("Description"),
    )
    copyright = MediaField(
        MP3StorageStyle("TCOP"),
        MP4StorageStyle("cprt"),
        StorageStyle("COPYRIGHT"),
        ASFStorageStyle("Copyright"),
    )
    bpm = MediaField(
        MP3StorageStyle("TBPM"),
        MP4StorageStyle("tmpo", as_type=int),
        StorageStyle("BPM"),
        ASFStorageStyle("WM/BeatsPerMinute"),
        out_type=int,
    )
    comp = MediaField(
        MP3StorageStyle("TCMP"),
        MP4BoolStorageStyle("cpil"),
        StorageStyle("COMPILATION"),
        ASFStorageStyle("WM/IsCompilation", as_type=bool),
        out_type=bool,
    )
    albumartist = MediaField(
        MP3StorageStyle("TPE2"),
        MP4StorageStyle("aART"),
        StorageStyle("ALBUM ARTIST"),
        StorageStyle("ALBUM_ARTIST"),
        StorageStyle("ALBUMARTIST"),
        ASFStorageStyle("WM/AlbumArtist"),
    )
    albumartists = ListMediaField(
        MP3ListDescStorageStyle(desc="ALBUMARTISTS"),
        MP3ListDescStorageStyle(desc="ALBUM_ARTISTS"),
        MP3ListDescStorageStyle(desc="ALBUM ARTISTS", read_only=True),
        MP4ListStorageStyle("----:com.apple.iTunes:ALBUMARTISTS"),
        MP4ListStorageStyle("----:com.apple.iTunes:ALBUM_ARTISTS"),
        MP4ListStorageStyle("----:com.apple.iTunes:ALBUM ARTISTS", read_only=True),
        ListStorageStyle("ALBUMARTISTS"),
        ListStorageStyle("ALBUM_ARTISTS"),
        ListStorageStyle("ALBUM ARTISTS", read_only=True),
        ASFStorageStyle("WM/AlbumArtists"),
    )
    albumtypes = ListMediaField(
        MP3ListDescStorageStyle("MusicBrainz Album Type", split_v23=True),
        MP4ListStorageStyle("----:com.apple.iTunes:MusicBrainz Album Type"),
        ListStorageStyle("RELEASETYPE"),
        ListStorageStyle("MUSICBRAINZ_ALBUMTYPE"),
        ASFStorageStyle("MusicBrainz/Album Type"),
    )
    albumtype = albumtypes.single_field()

    label = MediaField(
        MP3StorageStyle("TPUB"),
        MP3DescStorageStyle("LABEL"),
        MP4StorageStyle("----:com.apple.iTunes:LABEL"),
        MP4StorageStyle("----:com.apple.iTunes:publisher"),
        MP4StorageStyle("----:com.apple.iTunes:Label", read_only=True),
        StorageStyle("LABEL"),
        StorageStyle("PUBLISHER"),  # Traktor
        ASFStorageStyle("WM/Publisher"),
    )
    artist_sort = MediaField(
        MP3StorageStyle("TSOP"),
        MP4StorageStyle("soar"),
        StorageStyle("ARTISTSORT"),
        ASFStorageStyle("WM/ArtistSortOrder"),
    )
    albumartist_sort = MediaField(
        MP3StorageStyle("TSO2"),
        MP3DescStorageStyle("ALBUMARTISTSORT"),
        MP4StorageStyle("soaa"),
        StorageStyle("ALBUMARTISTSORT"),
        ASFStorageStyle("WM/AlbumArtistSortOrder"),
    )
    asin = MediaField(
        MP3DescStorageStyle("ASIN"),
        MP4StorageStyle("----:com.apple.iTunes:ASIN"),
        StorageStyle("ASIN"),
        ASFStorageStyle("MusicBrainz/ASIN"),
    )
    catalognums = ListMediaField(
        MP3ListDescStorageStyle("CATALOGNUMBER", split_v23=True),
        MP3ListDescStorageStyle("CATALOGID", read_only=True),
        MP3ListDescStorageStyle("DISCOGS_CATALOG", read_only=True),
        MP4ListStorageStyle("----:com.apple.iTunes:CATALOGNUMBER"),
        MP4ListStorageStyle("----:com.apple.iTunes:CATALOGID", read_only=True),
        MP4ListStorageStyle("----:com.apple.iTunes:DISCOGS_CATALOG", read_only=True),
        ListStorageStyle("CATALOGNUMBER"),
        ListStorageStyle("CATALOGID", read_only=True),
        ListStorageStyle("DISCOGS_CATALOG", read_only=True),
        ASFStorageStyle("WM/CatalogNo"),
        ASFStorageStyle("CATALOGID", read_only=True),
        ASFStorageStyle("DISCOGS_CATALOG", read_only=True),
    )
    catalognum = catalognums.single_field()

    barcode = MediaField(
        MP3DescStorageStyle("BARCODE"),
        MP4StorageStyle("----:com.apple.iTunes:BARCODE"),
        StorageStyle("BARCODE"),
        StorageStyle("UPC", read_only=True),
        StorageStyle("EAN/UPN", read_only=True),
        StorageStyle("EAN", read_only=True),
        StorageStyle("UPN", read_only=True),
        ASFStorageStyle("WM/Barcode"),
    )
    isrc = MediaField(
        MP3StorageStyle("TSRC"),
        MP4StorageStyle("----:com.apple.iTunes:ISRC"),
        StorageStyle("ISRC"),
        ASFStorageStyle("WM/ISRC"),
    )
    disctitle = MediaField(
        MP3StorageStyle("TSST"),
        MP4StorageStyle("----:com.apple.iTunes:DISCSUBTITLE"),
        StorageStyle("DISCSUBTITLE"),
        ASFStorageStyle("WM/SetSubTitle"),
    )
    encoder = MediaField(
        MP3StorageStyle("TENC"),
        MP4StorageStyle("\xa9too"),
        StorageStyle("ENCODEDBY"),
        StorageStyle("ENCODER"),
        ASFStorageStyle("WM/EncodedBy"),
    )
    script = MediaField(
        MP3DescStorageStyle("Script"),
        MP4StorageStyle("----:com.apple.iTunes:SCRIPT"),
        StorageStyle("SCRIPT"),
        ASFStorageStyle("WM/Script"),
    )
    languages = ListMediaField(
        MP3ListStorageStyle("TLAN"),
        MP4ListStorageStyle("----:com.apple.iTunes:LANGUAGE"),
        ListStorageStyle("LANGUAGE"),
        ASFStorageStyle("WM/Language"),
    )
    language = languages.single_field()

    country = MediaField(
        MP3DescStorageStyle("MusicBrainz Album Release Country"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Album Release Country"),
        StorageStyle("RELEASECOUNTRY"),
        ASFStorageStyle("MusicBrainz/Album Release Country"),
    )
    albumstatus = MediaField(
        MP3DescStorageStyle("MusicBrainz Album Status"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Album Status"),
        StorageStyle("RELEASESTATUS"),
        StorageStyle("MUSICBRAINZ_ALBUMSTATUS"),
        ASFStorageStyle("MusicBrainz/Album Status"),
    )
    media = MediaField(
        MP3StorageStyle("TMED"),
        MP3DescStorageStyle("MEDIA"),
        MP4StorageStyle("----:com.apple.iTunes:MEDIA"),
        StorageStyle("MEDIA"),
        ASFStorageStyle("WM/Media"),
    )
    albumdisambig = MediaField(
        # This tag mapping was invented for beets (not used by Picard, etc).
        MP3DescStorageStyle("MusicBrainz Album Comment"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Album Comment"),
        StorageStyle("MUSICBRAINZ_ALBUMCOMMENT"),
        ASFStorageStyle("MusicBrainz/Album Comment"),
    )

    # Release date.
    date = DateField(
        MP3StorageStyle("TDRC"),
        MP4StorageStyle("\xa9day"),
        StorageStyle("DATE"),
        ASFStorageStyle("WM/Year"),
        year=(StorageStyle("YEAR"),),
    )

    year = date.year_field()
    month = date.month_field()
    day = date.day_field()

    # *Original* release date.
    original_date = DateField(
        MP3StorageStyle("TDOR"),
        MP4StorageStyle("----:com.apple.iTunes:ORIGINAL YEAR"),
        MP4StorageStyle("----:com.apple.iTunes:ORIGINALDATE"),
        StorageStyle("ORIGINALDATE"),
        ASFStorageStyle("WM/OriginalReleaseYear"),
    )

    original_year = original_date.year_field()
    original_month = original_date.month_field()
    original_day = original_date.day_field()

    # Nonstandard metadata.
    artist_credit = MediaField(
        MP3DescStorageStyle("Artist Credit"),
        MP4StorageStyle("----:com.apple.iTunes:Artist Credit"),
        StorageStyle("ARTIST_CREDIT"),
        ASFStorageStyle("beets/Artist Credit"),
    )
    artists_credit = ListMediaField(
        MP3ListDescStorageStyle(desc="ARTISTS_CREDIT"),
        MP4ListStorageStyle("----:com.apple.iTunes:ARTISTS_CREDIT"),
        ListStorageStyle("ARTISTS_CREDIT"),
        ASFStorageStyle("beets/ArtistsCredit"),
    )
    artists_sort = ListMediaField(
        MP3ListDescStorageStyle(desc="ARTISTS_SORT"),
        MP4ListStorageStyle("----:com.apple.iTunes:ARTISTS_SORT"),
        ListStorageStyle("ARTISTS_SORT"),
        ASFStorageStyle("beets/ArtistsSort"),
    )
    albumartist_credit = MediaField(
        MP3DescStorageStyle("Album Artist Credit"),
        MP4StorageStyle("----:com.apple.iTunes:Album Artist Credit"),
        StorageStyle("ALBUMARTIST_CREDIT"),
        ASFStorageStyle("beets/Album Artist Credit"),
    )
    albumartists_credit = ListMediaField(
        MP3ListDescStorageStyle(desc="ALBUMARTISTS_CREDIT"),
        MP4ListStorageStyle("----:com.apple.iTunes:ALBUMARTISTS_CREDIT"),
        ListStorageStyle("ALBUMARTISTS_CREDIT"),
        ASFStorageStyle("beets/AlbumArtistsCredit"),
    )
    albumartists_sort = ListMediaField(
        MP3ListDescStorageStyle(desc="ALBUMARTISTS_SORT"),
        MP4ListStorageStyle("----:com.apple.iTunes:ALBUMARTISTS_SORT"),
        ListStorageStyle("ALBUMARTISTS_SORT"),
        ASFStorageStyle("beets/AlbumArtistsSort"),
    )
    remixers = ListMediaField(
        MP3ListStorageStyle("TPE4"),
        MP4ListStorageStyle("----:com.apple.iTunes:REMIXERS"),
        ListStorageStyle("REMIXER"),
        ASFStorageStyle("WM/REMIXERS"),
    )

    # Legacy album art field
    art = CoverArtField()

    # Image list
    images = ImageListField()

    # MusicBrainz IDs.
    mb_trackid = MediaField(
        MP3UFIDStorageStyle(owner="http://musicbrainz.org"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Track Id"),
        StorageStyle("MUSICBRAINZ_TRACKID"),
        ASFStorageStyle("MusicBrainz/Track Id"),
    )
    mb_releasetrackid = MediaField(
        MP3DescStorageStyle("MusicBrainz Release Track Id"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Release Track Id"),
        StorageStyle("MUSICBRAINZ_RELEASETRACKID"),
        ASFStorageStyle("MusicBrainz/Release Track Id"),
    )
    mb_workid = MediaField(
        MP3DescStorageStyle("MusicBrainz Work Id"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Work Id"),
        StorageStyle("MUSICBRAINZ_WORKID"),
        ASFStorageStyle("MusicBrainz/Work Id"),
    )
    mb_albumid = MediaField(
        MP3DescStorageStyle("MusicBrainz Album Id"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Album Id"),
        StorageStyle("MUSICBRAINZ_ALBUMID"),
        ASFStorageStyle("MusicBrainz/Album Id"),
    )
    mb_artistids = ListMediaField(
        MP3ListDescStorageStyle("MusicBrainz Artist Id", split_v23=True),
        MP4ListStorageStyle("----:com.apple.iTunes:MusicBrainz Artist Id"),
        ListStorageStyle("MUSICBRAINZ_ARTISTID"),
        ASFStorageStyle("MusicBrainz/Artist Id"),
    )
    mb_artistid = mb_artistids.single_field()

    mb_albumartistids = ListMediaField(
        MP3ListDescStorageStyle(
            "MusicBrainz Album Artist Id",
            split_v23=True,
        ),
        MP4ListStorageStyle(
            "----:com.apple.iTunes:MusicBrainz Album Artist Id",
        ),
        ListStorageStyle("MUSICBRAINZ_ALBUMARTISTID"),
        ASFStorageStyle("MusicBrainz/Album Artist Id"),
    )
    mb_albumartistid = mb_albumartistids.single_field()

    mb_releasegroupid = MediaField(
        MP3DescStorageStyle("MusicBrainz Release Group Id"),
        MP4StorageStyle("----:com.apple.iTunes:MusicBrainz Release Group Id"),
        StorageStyle("MUSICBRAINZ_RELEASEGROUPID"),
        ASFStorageStyle("MusicBrainz/Release Group Id"),
    )

    # Acoustid fields.
    acoustid_fingerprint = MediaField(
        MP3DescStorageStyle("Acoustid Fingerprint"),
        MP4StorageStyle("----:com.apple.iTunes:Acoustid Fingerprint"),
        StorageStyle("ACOUSTID_FINGERPRINT"),
        ASFStorageStyle("Acoustid/Fingerprint"),
    )
    acoustid_id = MediaField(
        MP3DescStorageStyle("Acoustid Id"),
        MP4StorageStyle("----:com.apple.iTunes:Acoustid Id"),
        StorageStyle("ACOUSTID_ID"),
        ASFStorageStyle("Acoustid/Id"),
    )

    # ReplayGain fields.
    rg_track_gain = MediaField(
        MP3DescStorageStyle("REPLAYGAIN_TRACK_GAIN", float_places=2, suffix=" dB"),
        MP3DescStorageStyle("replaygain_track_gain", float_places=2, suffix=" dB"),
        MP3SoundCheckStorageStyle(key="COMM", index=0, desc="iTunNORM", id3_lang="eng"),
        MP4StorageStyle(
            "----:com.apple.iTunes:replaygain_track_gain", float_places=2, suffix=" dB"
        ),
        MP4SoundCheckStorageStyle("----:com.apple.iTunes:iTunNORM", index=0),
        StorageStyle("REPLAYGAIN_TRACK_GAIN", float_places=2, suffix=" dB"),
        ASFStorageStyle("replaygain_track_gain", float_places=2, suffix=" dB"),
        out_type=float,
    )
    rg_album_gain = MediaField(
        MP3DescStorageStyle("REPLAYGAIN_ALBUM_GAIN", float_places=2, suffix=" dB"),
        MP3DescStorageStyle("replaygain_album_gain", float_places=2, suffix=" dB"),
        MP4StorageStyle(
            "----:com.apple.iTunes:replaygain_album_gain", float_places=2, suffix=" dB"
        ),
        StorageStyle("REPLAYGAIN_ALBUM_GAIN", float_places=2, suffix=" dB"),
        ASFStorageStyle("replaygain_album_gain", float_places=2, suffix=" dB"),
        out_type=float,
    )
    rg_track_peak = MediaField(
        MP3DescStorageStyle("REPLAYGAIN_TRACK_PEAK", float_places=6),
        MP3DescStorageStyle("replaygain_track_peak", float_places=6),
        MP3SoundCheckStorageStyle(key="COMM", index=1, desc="iTunNORM", id3_lang="eng"),
        MP4StorageStyle("----:com.apple.iTunes:replaygain_track_peak", float_places=6),
        MP4SoundCheckStorageStyle("----:com.apple.iTunes:iTunNORM", index=1),
        StorageStyle("REPLAYGAIN_TRACK_PEAK", float_places=6),
        ASFStorageStyle("replaygain_track_peak", float_places=6),
        out_type=float,
    )
    rg_album_peak = MediaField(
        MP3DescStorageStyle("REPLAYGAIN_ALBUM_PEAK", float_places=6),
        MP3DescStorageStyle("replaygain_album_peak", float_places=6),
        MP4StorageStyle("----:com.apple.iTunes:replaygain_album_peak", float_places=6),
        StorageStyle("REPLAYGAIN_ALBUM_PEAK", float_places=6),
        ASFStorageStyle("replaygain_album_peak", float_places=6),
        out_type=float,
    )

    # EBU R128 fields.
    r128_track_gain = QNumberField(
        8,
        MP3DescStorageStyle("R128_TRACK_GAIN"),
        MP4StorageStyle("----:com.apple.iTunes:R128_TRACK_GAIN"),
        StorageStyle("R128_TRACK_GAIN"),
        ASFStorageStyle("R128_TRACK_GAIN"),
    )
    r128_album_gain = QNumberField(
        8,
        MP3DescStorageStyle("R128_ALBUM_GAIN"),
        MP4StorageStyle("----:com.apple.iTunes:R128_ALBUM_GAIN"),
        StorageStyle("R128_ALBUM_GAIN"),
        ASFStorageStyle("R128_ALBUM_GAIN"),
    )

    initial_key = MediaField(
        MP3StorageStyle("TKEY"),
        MP4StorageStyle("----:com.apple.iTunes:initialkey"),
        StorageStyle("INITIALKEY"),
        ASFStorageStyle("INITIALKEY"),
    )

    @property
    def length(self):
        """The duration of the audio in seconds (a float)."""
        return self.mgfile.info.length

    @property
    def samplerate(self):
        """The audio's sample rate (an int)."""
        if hasattr(self.mgfile.info, "sample_rate"):
            return self.mgfile.info.sample_rate
        elif self.type == "opus":
            # Opus is always 48kHz internally.
            return 48000
        return 0

    @property
    def bitdepth(self):
        """The number of bits per sample in the audio encoding (an int).
        Only available for certain file formats (zero where
        unavailable).
        """
        if hasattr(self.mgfile.info, "bits_per_sample"):
            return self.mgfile.info.bits_per_sample
        return 0

    @property
    def channels(self):
        """The number of channels in the audio (an int)."""
        if hasattr(self.mgfile.info, "channels"):
            return self.mgfile.info.channels
        return 0

    @property
    def bitrate(self):
        """The number of bits per seconds used in the audio coding (an
        int). If this is provided explicitly by the compressed file
        format, this is a precise reflection of the encoding. Otherwise,
        it is estimated from the on-disk file size. In this case, some
        imprecision is possible because the file header is incorporated
        in the file size.
        """
        if hasattr(self.mgfile.info, "bitrate") and self.mgfile.info.bitrate:
            # Many formats provide it explicitly.
            return self.mgfile.info.bitrate
        else:
            # Otherwise, we calculate bitrate from the file size. (This
            # is the case for all of the lossless formats.)
            if not self.length:
                # Avoid division by zero if length is not available.
                return 0
            return int(self.filesize * 8 / self.length)

    @property
    def bitrate_mode(self):
        """The mode of the bitrate used in the audio coding
        (a string, eg. "CBR", "VBR" or "ABR").
        Only available for the MP3 file format (empty where unavailable).
        """
        if hasattr(self.mgfile.info, "bitrate_mode"):
            return {
                mutagen.mp3.BitrateMode.CBR: "CBR",
                mutagen.mp3.BitrateMode.VBR: "VBR",
                mutagen.mp3.BitrateMode.ABR: "ABR",
            }.get(self.mgfile.info.bitrate_mode, "")
        else:
            return ""

    @property
    def encoder_info(self):
        """The name and/or version of the encoder used
        (a string, eg. "LAME 3.97.0").
        Only available for some formats (empty where unavailable).
        """
        if hasattr(self.mgfile.info, "encoder_info"):
            return self.mgfile.info.encoder_info
        else:
            return ""

    @property
    def encoder_settings(self):
        """A guess of the settings used for the encoder (a string, eg. "-V2").
        Only available for the MP3 file format (empty where unavailable).
        """
        if hasattr(self.mgfile.info, "encoder_settings"):
            return self.mgfile.info.encoder_settings
        else:
            return ""

    @property
    def format(self):
        """A string describing the file format/codec."""
        return TYPES[self.type]



import mutagen
import mutagen.id3

from mediafile.utils import Image

from .base import ListStorageStyle, SoundCheckStorageStyleMixin, StorageStyle


class MP3StorageStyle(StorageStyle):
    """Store data in ID3 frames."""

    formats = ["MP3", "AIFF", "DSF", "WAVE"]

    def __init__(self, key, id3_lang=None, **kwargs):
        """Create a new ID3 storage style. `id3_lang` is the value for
        the language field of newly created frames.
        """
        self.id3_lang = id3_lang
        super().__init__(key, **kwargs)

    def fetch(self, mutagen_file):
        try:
            return mutagen_file[self.key].text[0]
        except (KeyError, IndexError):
            return None

    def store(self, mutagen_file, value):
        frame = mutagen.id3.Frames[self.key](encoding=3, text=[value])
        mutagen_file.tags.setall(self.key, [frame])


class MP3PeopleStorageStyle(MP3StorageStyle):
    """Store list of people in ID3 frames."""

    def __init__(self, key, involvement="", **kwargs):
        self.involvement = involvement
        super().__init__(key, **kwargs)

    def _matching_people(self, mutagen_file):
        values = []
        for frame in mutagen_file.tags.getall(self.key):
            for pair in frame.people:
                if pair[0].lower() == self.involvement.lower() and len(pair) > 1:
                    values.append(pair[1])
        return values

    def _other_people(self, mutagen_file):
        values = []
        for frame in mutagen_file.tags.getall(self.key):
            for pair in frame.people:
                if pair[0].lower() != self.involvement.lower():
                    values.append(pair)
        return values

    def _store_people(self, mutagen_file, values):
        people = self._other_people(mutagen_file)
        people.extend([[self.involvement, value] for value in values])

        if people:
            frame = mutagen.id3.Frames[self.key](
                encoding=mutagen.id3._specs.Encoding.UTF8,
                people=people,
            )
            mutagen_file.tags.setall(self.key, [frame])
        else:
            mutagen_file.tags.delall(self.key)

    def store(self, mutagen_file, value):
        self._store_people(mutagen_file, [value])

    def fetch(self, mutagen_file):
        try:
            return self._matching_people(mutagen_file)[0]
        except IndexError:
            return None

    def delete(self, mutagen_file):
        self._store_people(mutagen_file, [])


class MP3ListPeopleStorageStyle(ListStorageStyle, MP3PeopleStorageStyle):
    """Store multiple values for one involvement in ID3 people frames."""

    def fetch(self, mutagen_file):
        return self._matching_people(mutagen_file)

    def store(self, mutagen_file, values):
        self._store_people(mutagen_file, values)


class MP3ListStorageStyle(ListStorageStyle, MP3StorageStyle):
    """Store lists of data in multiple ID3 frames."""

    def fetch(self, mutagen_file):
        try:
            return mutagen_file[self.key].text
        except KeyError:
            return []

    def store(self, mutagen_file, values):
        frame = mutagen.id3.Frames[self.key](encoding=3, text=values)
        mutagen_file.tags.setall(self.key, [frame])


class MP3UFIDStorageStyle(MP3StorageStyle):
    """Store string data in a UFID ID3 frame with a particular owner."""

    def __init__(self, owner, **kwargs):
        self.owner = owner
        super().__init__("UFID:" + owner, **kwargs)

    def fetch(self, mutagen_file):
        try:
            return mutagen_file[self.key].data
        except KeyError:
            return None

    def store(self, mutagen_file, value):
        # This field type stores text data as encoded data.
        assert isinstance(value, str)
        value = value.encode("utf-8")

        frames = mutagen_file.tags.getall(self.key)
        for frame in frames:
            # Replace existing frame data.
            if frame.owner == self.owner:
                frame.data = value
        else:
            # New frame.
            frame = mutagen.id3._frames.UFID(owner=self.owner, data=value)
            mutagen_file.tags.setall(self.key, [frame])


class MP3DescStorageStyle(MP3StorageStyle):
    """Store data in a TXXX (or similar) ID3 frame. The frame is
    selected based its ``desc`` field.
    ``attr`` allows to specify name of data accessor property in the frame.
    Most of frames use `text`.
    ``multispec`` specifies if frame data is ``mutagen.id3.MultiSpec``
    which means that the data is being packed in the list.
    """

    def __init__(self, desc="", key="TXXX", attr="text", multispec=True, **kwargs):
        assert isinstance(desc, str)
        self.description = desc
        self.attr = attr
        self.multispec = multispec
        super().__init__(key=key, **kwargs)

    def store(self, mutagen_file, value):
        frames = mutagen_file.tags.getall(self.key)
        if self.multispec:
            value = [value]

        # Try modifying in place.
        found = False
        for frame in frames:
            if frame.desc.lower() == self.description.lower():
                setattr(frame, self.attr, value)
                frame.encoding = mutagen.id3._specs.Encoding.UTF8
                found = True

        # Try creating a new frame.
        if not found:
            frame = mutagen.id3.Frames[self.key](
                desc=self.description,
                encoding=mutagen.id3._specs.Encoding.UTF8,
                **{self.attr: value},
            )
            if self.id3_lang:
                frame.lang = self.id3_lang
            mutagen_file.tags.add(frame)

    def fetch(self, mutagen_file):
        for frame in mutagen_file.tags.getall(self.key):
            if frame.desc.lower() == self.description.lower():
                if not self.multispec:
                    return getattr(frame, self.attr)
                try:
                    return getattr(frame, self.attr)[0]
                except IndexError:
                    return None

    def delete(self, mutagen_file):
        found_frame = None
        for frame in mutagen_file.tags.getall(self.key):
            if frame.desc.lower() == self.description.lower():
                found_frame = frame
                break
        if found_frame is not None:
            del mutagen_file[frame.HashKey]


class MP3ListDescStorageStyle(MP3DescStorageStyle, ListStorageStyle):
    def __init__(self, desc="", key="TXXX", split_v23=False, **kwargs):
        self.split_v23 = split_v23
        super().__init__(desc=desc, key=key, **kwargs)

    def fetch(self, mutagen_file):
        for frame in mutagen_file.tags.getall(self.key):
            if frame.desc.lower() == self.description.lower():
                if mutagen_file.tags.version == (2, 3, 0) and self.split_v23:
                    return sum((el.split("/") for el in frame.text), [])
                else:
                    return frame.text
        return []

    def store(self, mutagen_file, values):
        self.delete(mutagen_file)
        frame = mutagen.id3.Frames[self.key](
            desc=self.description,
            text=values,
            encoding=mutagen.id3._specs.Encoding.UTF8,
        )
        if self.id3_lang:
            frame.lang = self.id3_lang
        mutagen_file.tags.add(frame)


class MP3SlashPackStorageStyle(MP3StorageStyle):
    """Store value as part of pair that is serialized as a slash-
    separated string.
    """

    def __init__(self, key, pack_pos=0, **kwargs):
        super().__init__(key, **kwargs)
        self.pack_pos = pack_pos

    def _fetch_unpacked(self, mutagen_file):
        data = self.fetch(mutagen_file)
        if data:
            items = str(data).split("/")
        else:
            items = []
        packing_length = 2
        return list(items) + [None] * (packing_length - len(items))

    def get(self, mutagen_file):
        return self._fetch_unpacked(mutagen_file)[self.pack_pos]

    def set(self, mutagen_file, value):
        items = self._fetch_unpacked(mutagen_file)
        items[self.pack_pos] = value
        if items[0] is None:
            items[0] = ""
        if items[1] is None:
            items.pop()  # Do not store last value
        self.store(mutagen_file, "/".join(map(str, items)))

    def delete(self, mutagen_file):
        if self.pack_pos == 0:
            super().delete(mutagen_file)
        else:
            self.set(mutagen_file, None)


class MP3ImageStorageStyle(ListStorageStyle, MP3StorageStyle):
    """Converts between APIC frames and ``Image`` instances.

    The `get_list` method inherited from ``ListStorageStyle`` returns a
    list of ``Image``s. Similarly, the `set_list` method accepts a
    list of ``Image``s as its ``values`` argument.
    """

    def __init__(self):
        super().__init__(key="APIC")
        self.as_type = bytes

    def deserialize(self, apic_frame):
        """Convert APIC frame into Image."""
        return Image(data=apic_frame.data, desc=apic_frame.desc, type=apic_frame.type)

    def fetch(self, mutagen_file):
        return mutagen_file.tags.getall(self.key)

    def store(self, mutagen_file, frames):
        mutagen_file.tags.setall(self.key, frames)

    def delete(self, mutagen_file):
        mutagen_file.tags.delall(self.key)

    def serialize(self, image):
        """Return an APIC frame populated with data from ``image``."""
        assert isinstance(image, Image)
        frame = mutagen.id3.Frames[self.key]()
        frame.data = image.data
        frame.mime = image.mime_type
        frame.desc = image.desc or ""

        # For compatibility with OS X/iTunes prefer latin-1 if possible.
        # See issue #899
        try:
            frame.desc.encode("latin-1")
        except UnicodeEncodeError:
            frame.encoding = mutagen.id3._specs.Encoding.UTF16
        else:
            frame.encoding = mutagen.id3._specs.Encoding.LATIN1

        frame.type = image.type_index
        return frame


class MP3SoundCheckStorageStyle(SoundCheckStorageStyleMixin, MP3DescStorageStyle):
    def __init__(self, index=0, **kwargs):
        super().__init__(**kwargs)
        self.index = index
